//EE4840 - TerrorMouse MIDI Synthesizer - main program
//by Ron Weiss

#include "xbasic_types.h"
#include "xio.h"

#include "synth.h"
#include "wglookup.h"
#include "fmlookup.h"

#define UART_ADDR   0xFEFF0100
#define UART_RXFIFO 0xFEFF0100
#define UART_TXFIFO 0xFEFF0104
#define UART_STATUS 0xFEFF0108
#define UART_CTRL   0xFEFF010C

main()
{
    unsigned int  x;
    unsigned char byte = 0;
    unsigned char status = 0, databytes = 0;
    unsigned char data[2] = {0, 0};
    unsigned char key = 0, vel = 0;
    unsigned char synth_sel = 0, fm_mod = 0, fm_octave = 0;

    unsigned char wgs[WG_VOICES] = {0, 0, 0, 0, 0, 0}; 
    unsigned char fms[FM_VOICES] = {0, 0, 0, 0, 0, 0}; 

    /*
     * Initialization
     */

    //shut off LEDS
    for(x = 0; x < 256; x += 4)
        XIo_Out32(LED_ADDR+x, 0);

    //zero synth inputs
    for(x = 0; x < 256; x += 4)
        XIo_Out32(SYNTH_ADDR+x, 0);

    //turn on waveguide by default
    synth_sel = SEL_WG;
    XIo_Out8(SYNTH_SEL, synth_sel);

    //set up uartlite - disable interrupts/reset FIFOs
    XIo_Out32(UART_CTRL, 0);


    /*
     * The main loop - where the magic happens
     */

    for(;;)
    {
        //wait for new byte in UART_RXFIFO
        if(XIo_In32(UART_STATUS) & 0x01)
        {
            byte = XIo_In32(UART_RXFIFO);

            //first bit of status message is 1
            if(byte >> 7)
            {
                status = byte >> 4;
                data[0] = 0;
                data[1] = 0;
                databytes = 0;
            }
            else
            {
                data[databytes] = byte;
                databytes++;
            }

            //show something on the leds (key number or program number for our
            //purposes)
            XIo_Out8(LED_ADDR+8, data[0]);        

            if(databytes == 2) //2 data byte messages
            {
                key = data[0];
                vel = data[1];

                switch(synth_sel)
                {
                case SEL_FM:
                    switch(status)
                    {
                    case MIDI_NOTEOFF:
                        //turn off FM voice corresponding to key
                        for(x=0; x <= FM_VOICES-1; x++)
                        {
                            if(fms[x] == key)
                            {
                                fms[x] = 0;
                                fm_off(x);
                                break;
                            }
                        }
                        break;
                    case MIDI_NOTEON:
                        if(vel != 0) //note on
                        {
                            //find unused fm voice
                            for(x=0; x <= FM_VOICES-1; x++)
                            {
                                if(fms[x] == 0)
                                {
                                    fms[x] = key;
                                    fm_on(x, fm_notes[key - fm_octave]);
                                    break;
                                }
                            }
                        }
                        else //note off if we get note on message with vel=0
                        {
                            //turn off fm voice corresponding to key
                            for(x=0; x <= FM_VOICES-1; x++)
                            {
                                if(fms[x] == key)
                                {
                                    fms[x] = 0;
                                    fm_off(x);
                                    break;
                                }
                            }
                        }
                        break;
                    } //switch status
                    break;
                case SEL_WG:
                    switch(status)
                    {
                    case MIDI_NOTEOFF: 
                        //turn off waveguide corresponding to key
                        for(x=0; x <= WG_VOICES-1; x++)
                        {
                            if(wgs[x] == key)
                            {
                                wgs[x] = 0;
                                wg_off(x);
                                break;
                            }
                        }
                        break;
                    case MIDI_NOTEON: 
                        if(vel != 0 && key > 43 && key < 95) //note on
                        {
                            //find empty waveguide
                            for(x=0; x <= WG_VOICES-1; x++)
                            {
                                if(wgs[x] == 0)
                                {
                                    wgs[x] =  key;
                                    wg_on(x, wg_notes[key]);
                                    break;
                                }
                            }
                        }
                        else //note off if we get note on message with velocity of zero
                        {
                            //turn off waveguide corresponding to key
                            for(x=0; x <= WG_VOICES-1; x++)
                            {
                                if(wgs[x] == key)
                                {
                                    wgs[x] = 0;
                                    wg_off(x);
                                    break;
                                }
                            }
                        }        
                        break;
                    } //switch status
                } //switch synth_sel
                
                //reset registers and wait for next message
                databytes = 0;
            }            
            //we're only interpretting one message with only 1 data byte
            else if(status == MIDI_PROG && databytes == 1)
            {
                if(data[0] == 0) //select waveguide
                    synth_sel = SEL_WG;
                else if(data[0] <= FM_PATCHES)
                {
                    synth_sel = SEL_FM;
                    fm_mod    = data[0] - 1;
                    
                    //compensate for patches with large modulating frequency - shift down an octave
                    if(data[0] == 4 || data[0] == 6 || data[0] == 8 || data[0] == 10)
                        fm_octave = 12;
                    else
                        fm_octave = 0;
                }
                XIo_Out8(SYNTH_SEL, synth_sel);
                XIo_Out8(FM_MOD, fm_mod);

                //turn off stuck notes:
                for(x = 0; x <= 5; x++)
                {
                    wgs[x] = 0;
                    wg_off(x);
                    fms[x] = 0;
                    fm_off(x);
                }                

                //reset registers and wait for next message
                databytes = 0;
            }
        } //if new byte ready
    } //for
}
