-- CSEEE4830 - TerrorMouse
-- by Ron Weiss
library ieee;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_SIGNED.all;

-------------------------------------------------------------------------------
-- entity
-------------------------------------------------------------------------------

entity OPB_synth is  --USER-- change entity name
  generic
  (
    C_BASEADDR         : std_logic_vector(0 to 31) := X"FFFFFFFF";
    C_HIGHADDR         : std_logic_vector(0 to 31) := X"00000000";
    C_OPB_AWIDTH       : integer                   := 32;
    C_OPB_DWIDTH       : integer                   := 32;
    C_FAMILY           : string                    := "spartan2e"
  );
  port
  (
    --Required OPB bus ports, do not add to or delete
    OPB_ABus     : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE       : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_Clk      : in  std_logic;
    OPB_DBus     : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW      : in  std_logic;
    OPB_Rst      : in  std_logic;
    OPB_select   : in  std_logic;
    OPB_seqAddr  : in  std_logic;
    Sln_DBus     : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck   : out std_logic;
    Sln_retry    : out std_logic;
    Sln_toutSup  : out std_logic;
    Sln_xferAck  : out std_logic;

    --USER-- add user I/Os to port list
    FPGA_clk : in  std_logic; -- FPGA clock
    AU_CSN_N : out std_logic;
    AU_BCLK  : out std_logic;
    AU_MCLK  : out std_logic;
    AU_LRCK  : out std_logic;
    AU_SDTI  : out std_logic
  );
end entity OPB_synth; --USER-- change entity name

-------------------------------------------------------------------------------
-- architecture
-------------------------------------------------------------------------------

architecture imp of OPB_synth is --USER-- change entity name

  component audio_out
    port (
      clk : in std_logic;
      rst : in std_logic;
      
      sample_in : in std_logic_vector(15 downto 0);
      
      AU_CSN_N : out std_logic;
      AU_BCLK  : out std_logic;
      AU_MCLK  : out std_logic;
      AU_LRCK  : out std_logic;
      AU_SDTI  : out std_logic
      );    
  end component;

  component fm_synth
    port (
      clk        : in std_logic;
      sample_out : out std_logic_vector(15 downto 0);
      reset      : in std_logic;
      
      --offset inputs
      theta1_bres_in   : in std_logic_vector(31 downto 0);
      theta2_bres_in   : in std_logic_vector(31 downto 0);
      theta3_bres_in   : in std_logic_vector(31 downto 0);
      theta4_bres_in   : in std_logic_vector(31 downto 0);
      theta5_bres_in   : in std_logic_vector(31 downto 0);
      theta6_bres_in   : in std_logic_vector(31 downto 0);

      --enables
      theta1_en        : in std_logic;
      theta2_en        : in std_logic;   
      theta3_en        : in std_logic;
      theta4_en        : in std_logic;
      theta5_en        : in std_logic;   
      theta6_en        : in std_logic;
    
      --c/m ratio input
      cToM             : in std_logic_vector(7 downto 0)
      );        
  end component;

  component waveguide
    port (
      clk        : in  std_logic;
      enable     : in  std_logic;
      reset      : in  std_logic;
      length     : in  std_logic_vector(7 downto 0);   -- length of delay line
      sample_out : out std_logic_vector(15 downto 0)
    );
  end component;

  signal addr  : std_logic_vector(0 to 7);  -- local address
  signal wdata : std_logic_vector(0 to 7);
  signal rnw   : std_logic;
  signal cs, xfer : std_logic;

  signal byte_data : std_logic_vector(7 downto 0);
  signal word_data : std_logic_vector(31 downto 0);
  
  -- waveguide synth signals (6 voices)
  signal wg1_en  : std_logic                     := '1';
  signal wg1_rst : std_logic                     := '0';
  signal wg1_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg1_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg2_en  : std_logic                     := '1';
  signal wg2_rst : std_logic                     := '0';
  signal wg2_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg2_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg3_en  : std_logic                     := '1';
  signal wg3_rst : std_logic                     := '0';
  signal wg3_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg3_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg4_en  : std_logic                     := '1';
  signal wg4_rst : std_logic                     := '0';
  signal wg4_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg4_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg5_en  : std_logic                     := '1';
  signal wg5_rst : std_logic                     := '0';
  signal wg5_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg5_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg6_en  : std_logic                     := '1';  
  signal wg6_rst : std_logic                     := '0';
  signal wg6_len : std_logic_vector(7 downto 0)  := "00000000";
  signal wg6_out : std_logic_vector(15 downto 0) := "0000000000000000";
  signal wg_out  : std_logic_vector(18 downto 0) := "0000000000000000000";

  -- fm synth signals
  signal fm1_en    : std_logic := '0';
  signal fm1_theta : std_logic_vector(31 downto 0);
  signal fm2_en    : std_logic := '0';
  signal fm2_theta : std_logic_vector(31 downto 0);
  signal fm3_en    : std_logic := '0';
  signal fm3_theta : std_logic_vector(31 downto 0);
  signal fm4_en    : std_logic := '0';
  signal fm4_theta : std_logic_vector(31 downto 0);
  signal fm5_en    : std_logic := '0';
  signal fm5_theta : std_logic_vector(31 downto 0);
  signal fm6_en    : std_logic := '0';
  signal fm6_theta : std_logic_vector(31 downto 0);
  signal fm_mod   : std_logic_vector(7 downto 0);  -- which FM patch
  signal fm_out   : std_logic_vector(15 downto 0);

  signal synth_sel : std_logic := '0';   -- synth selector (0 = fm, 1 = wg);
  signal synth_out : std_logic_vector(15 downto 0);
  
begin
  process (OPB_select, OPB_ABus)
  begin
   if(OPB_select='1' and OPB_ABus(0 to 23)=C_BASEADDR(0 to 23)) then
     cs <= '1';
   else
     cs <= '0';
   end if;
  end process;

  -- swap bit order of data
  byte_data(7 downto 0)  <= wdata(0 to 7);
  
  process (OPB_Clk)
  begin
    if OPB_Clk'event and OPB_Clk = '1' then
      rnw   <= OPB_RNW;    
      addr  <= OPB_ABus(24 to 31);
      wdata <= OPB_DBus(0 to 7);
      xfer  <= cs and not xfer;

      word_data(31 downto 0) <= OPB_DBus(0 to 31);  
      
      if(xfer='1' and rnw='0') then
        case addr is
          when X"10" => wg1_en    <= word_data(0); 
          when X"11" => wg1_rst   <= word_data(0); 
          when X"12" => wg1_len   <= byte_data;    
          when X"20" => wg2_en    <= byte_data(0); 
          when X"21" => wg2_rst   <= byte_data(0); 
          when X"22" => wg2_len   <= byte_data;    
          when X"30" => wg3_en    <= byte_data(0); 
          when X"31" => wg3_rst   <= byte_data(0); 
          when X"32" => wg3_len   <= byte_data;    
          when X"40" => wg4_en    <= byte_data(0); 
          when X"41" => wg4_rst   <= byte_data(0); 
          when X"42" => wg4_len   <= byte_data;    
          when X"50" => wg5_en    <= byte_data(0); 
          when X"51" => wg5_rst   <= byte_data(0); 
          when X"52" => wg5_len   <= byte_data;    
          when X"60" => wg6_en    <= byte_data(0); 
          when X"61" => wg6_rst   <= byte_data(0); 
          when X"62" => wg6_len   <= byte_data;    
          when X"70" => fm1_en    <= byte_data(0); 
          when X"72" => fm1_theta <= word_data;    
          when X"80" => fm2_en    <= word_data(0); 
          when X"82" => fm2_theta <= word_data;    
          when X"90" => fm3_en    <= word_data(0); 
          when X"92" => fm3_theta <= word_data;    
          when X"a0" => fm4_en    <= word_data(0); 
          when X"a2" => fm4_theta <= word_data;   
          when X"b0" => fm5_en    <= word_data(0);
          when X"b2" => fm5_theta <= word_data;   
          when X"c0" => fm6_en    <= word_data(0);
          when X"c2" => fm6_theta <= word_data;   
          when X"f0" => fm_mod    <= byte_data;   
          when X"ff" => synth_sel <= byte_data(0);
          when others => null;
        end case;
      end if;     
    end if;
  end process;

  Sln_xferAck <= xfer;

  -- which synth are we using?
  process(OPB_Clk)
  begin
    if OPB_Clk'event and OPB_Clk = '1' then
      if synth_sel ='0' then
        synth_out <= fm_out;
      else
        wg_out <= conv_std_logic_vector((
          conv_integer(wg1_out) + conv_integer(wg2_out) +
          conv_integer(wg3_out) + conv_integer(wg4_out) +
          conv_integer(wg5_out) + conv_integer(wg6_out)), 19);

        synth_out <= wg_out(18 downto 3);
      end if;
    end if;
  end process;

  fm : fm_synth
    port map (
      clk        => FPGA_clk,
      sample_out => fm_out,
      reset      => '0',
      
      theta1_bres_in => fm1_theta,
      theta2_bres_in => fm2_theta,
      theta3_bres_in => fm3_theta,
      theta4_bres_in => fm4_theta,
      theta5_bres_in => fm5_theta,
      theta6_bres_in => fm6_theta,

      theta1_en      => fm1_en,
      theta2_en      => fm2_en,
      theta3_en      => fm3_en,
      theta4_en      => fm4_en,
      theta5_en      => fm5_en,
      theta6_en      => fm6_en,
    
      cToM           => fm_mod
   );

  wg1 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg1_en,
      reset      => wg1_rst,
      length     => wg1_len,
      sample_out => wg1_out
      );

  wg2 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg2_en,
      reset      => wg2_rst,
      length     => wg2_len,
      sample_out => wg2_out
      );

  wg3 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg3_en,
      reset      => wg3_rst,
      length     => wg3_len,
      sample_out => wg3_out
      );

  wg4 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg4_en,
      reset      => wg4_rst,
      length     => wg4_len,
      sample_out => wg4_out
      );

  wg5 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg5_en,
      reset      => wg5_rst,
      length     => wg5_len,
      sample_out => wg5_out
      );

  wg6 : waveguide
    port map (
      clk        => FPGA_clk,
      enable     => wg6_en,
      reset      => wg6_rst,
      length     => wg6_len,
      sample_out => wg6_out
      );

  audo : audio_out
    port map (
      CLK => FPGA_clk,
      RST => OPB_Rst,
      sample_in => synth_out,
        
      AU_CSN_N => AU_CSN_N,
      AU_BCLK  => AU_BCLK,
      AU_MCLK  => AU_MCLK,
      AU_LRCK  => AU_LRCK,
      AU_SDTI  => AU_SDTI
    );  
end architecture imp;
